#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Gevent pool/group alike: make asyncio easier to use

Usage::

>>> @asyncio.coroutine
>>> def f(url):
...     r = yield from aiohttp.request('get', url)
...     content = yield from r.read()
...     print('{}: {}'.format(url, content[:80]))

>>> g = Group()
>>> g.async(f('http://www.baidu.com'))
>>> g.async(f('http://www.sina.com.cn'))
>>> g.join()

>>> # limit the concurrent coroutines to 3
>>> p = Pool(3)
>>> for _ in range(10):
...     p.async(f('http://www.baidu.com'))
>>> p.join()
"""
import asyncio


class Group(object):

    def __init__(self, loop=None):
        try:
            self.loop = loop or asyncio.get_event_loop()
            if self.loop._running:
                raise NotImplementedError("Cannot use aioutils in "
                                          "asynchroneous environment")
        except:
            self.loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self.loop)
        self._prepare()

    def _prepare(self):
        self.counter = 0
        self.task_waiter = asyncio.futures.Future(loop=self.loop)

    def spawn(self, coro_or_future):
        self.counter += 1
        task = asyncio.async(coro_or_future)
        task.add_done_callback(self._on_completion)
        return task

    async = spawn

    def _on_completion(self, f):
        self.counter -= 1
        f.remove_done_callback(self._on_completion)
        if self.counter <= 0:
            if not self.task_waiter.done():
                self.task_waiter.set_result(None)

    def join(self):
        def _on_waiter(f):
            self.loop.stop()
            self._prepare()

        self.task_waiter.add_done_callback(_on_waiter)

        # expect the loops to be stop and start multiple times
        while self.counter > 0:
            if not self.loop._running:
                self.loop.run_forever()


class Pool(Group):

    def __init__(self, pool_size, loop=None):
        self.sem = asyncio.Semaphore(pool_size, loop=loop)
        super(Pool, self).__init__(loop)

    def spawn(self, coro):
        assert asyncio.iscoroutine(coro), 'pool only accepts coroutine'

        @asyncio.coroutine
        def _limit_coro():
            with (yield from self.sem):
                return (yield from coro)

        self.counter += 1
        task = asyncio.async(_limit_coro())
        task.add_done_callback(self._on_completion)
        return task

    async = spawn



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import random
import inspect
import asyncio

from aioutils import Bag, OrderedBag, Group

def test_bag():
    chars = 'abcdefg'
    def g():
        b = Bag()
        @asyncio.coroutine
        def f(c):
            yield from asyncio.sleep(random.random()/10)
            b.put(c)
        def schedule():
            for c in chars:
                b.spawn(f(c))
            b.join()
        b.schedule(schedule)
        yield from b.yielder()

    chars2 = g()
    assert inspect.isgenerator(chars2)
    chars2 = list(chars2)
    assert set(chars) == set(chars2)


def test_orderedbag():
    chars = 'abcdefg'
    def g():
        b = OrderedBag(Group())
        @asyncio.coroutine
        def f(c):
            yield from asyncio.sleep(random.random()*0.1)
            b.put(c)
        def schedule():
            for c in chars:
                b.spawn(f(c))
            b.join()
        b.schedule(schedule)
        yield from b.yielder()

    chars2 = g()
    assert inspect.isgenerator(chars2)
    chars2 = list(chars2)
    for c1, c2 in zip(chars, chars2):
        assert c1 == c2


if __name__ == '__main__':
    test_bag()
    test_orderedbag()



from .pool import Pool, Group
from .bag import Bag, OrderedBag
from .yielder import Yielder, OrderedYielder, yielding, ordered_yielding

__all__ = ['Pool', 'Group', 'Bag', 'OrderedBag',
           'Yielder', 'OrderedYielder', 'yielding', 'ordered_yielding']
__version__ = '0.3.4'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import time
import random
import asyncio
import threading
from aioutils import Group, Yielder

@asyncio.coroutine
def f(c):
    yield from asyncio.sleep(random.random()*0.02)
    return c


def test_group_threading():
    """ Ensure that Pool and Group are thread-safe """
    stopall = False
    def t():
        while not stopall:
            g = Group()
            for i in range(5):
                g.spawn(f(i))

            g.join()

            time.sleep(random.random()*0.02)

    tasks = [threading.Thread(target=t) for _ in range(5)]
    for task in tasks: task.daemon = True
    for task in tasks: task.start()
    time.sleep(0.2)
    stopall = True
    for task in tasks: task.join()
    assert asyncio.Task.all_tasks() == set(), asyncio.Task.all_tasks()


def test_yielder_threading():
    """ Ensure Yielder are thread safe """
    stopall = False
    chars = 'abcdefg'

    def gen_func():
        y = Yielder()

        for c in chars:
            y.spawn(f(c))

        yield from y.yielding()

    def t():
        while not stopall:
            chars2 = list(gen_func())
            assert set(chars2) == set(chars)

            time.sleep(random.random()*0.02)

    tasks = [threading.Thread(target=t) for _ in range(5)]
    for task in tasks: task.daemon = True
    for task in tasks: task.start()
    time.sleep(0.2)
    stopall = True
    for task in tasks: task.join()
    assert asyncio.Task.all_tasks() == set(), asyncio.Task.all_tasks()


def test_mixed():
    """ Ensure mixed usage are thread safe """
    chars = 'abcdefg'
    stopall = False

    def f1():
        y = Yielder()
        for c in chars:
            y.spawn(f(c))

        return list(y.yielding())

    def f2():
        g = Group()
        for c in chars:
            g.spawn(f(c))

        g.join()

    def t():
        while not stopall:
            f = random.choice([f1, f2])
            r = f()
            if f == f1:
                assert set(r) == set(chars)
            time.sleep(random.random()*0.02)

    tasks = [threading.Thread(target=t) for _ in range(5)]
    for task in tasks: task.daemon = True
    for task in tasks: task.start()
    time.sleep(0.2)
    stopall = True
    for task in tasks: task.join()
    assert asyncio.Task.all_tasks() == set(), asyncio.Task.all_tasks()


if __name__ == '__main__':
    test_group_threading()
    test_yielder_threading()
    test_mixed()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup

from aioutils import __version__

with open('HISTORY.md') as f:
    history = f.read()

setup(
    name='aioutils',
    version=__version__,
    description='Python3 Asyncio Utils',
    long_description= history,
    author='Jingchao Hu',
    author_email='jingchaohu@gmail.com',
    url='https://github.com/observerss/aioutils',
    packages=['aioutils'],
    license='Apache 2.0',
    zip_safe=True,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4'
        ],
    )
